/**
    Hydraulic Sandbox
    Version 1.50
    Martin Hebel - developer
    Karen Johnson - SME
    2022
    
    support@selmaware.com   - best email address that I'll see the Email
    
    This software is copyrighted by Martin Hebel.
    You may modify it for your own or your organizations use,
    but public distribution is not permitted, in whole or in part,
    without written permission.
    
    Please don't contact me with code questions, I'd need to decifer it myself again in most cases
 
    Some code may be older unused code for the final operation.
**/

/** General Component Object Conventions
- Components are all instanced on load, there is a set number of each.
- All components are generic, but have an image, component number and ID (type) number assigned to them.
- For example, the 5 variable pumps are components 10-14 with an ID of 3
- Each component can have up to 4 ports, they are numbered as component number *4+0 for first, +1 for second, +2 and +3 for third, forth
  such that variable pump 11 would have port number of 44 to 47, pump 12 would have ports 48 to 52, etc.
- Hoses are created with their endpoints assigned to a port, which in turn identified a component by dividing its number by 4.
- Connection verification is done by way of checking each hose and verifying the end port connections 
  (and thus the component types) make a legal connection.
- Tees use recursive checks to verify each successive tee connection to a component is legal
**/

String ver = "1.50";
boolean terminal = false;                      // sets the mode, with enable switch or not. True locks when turned on.
boolean toggleFlag = false;                    // Initial state of mode switch

String date = "Jan 10, 2022";

Components[] component = new Components[285];  //Class instance for all the components 
Hoses[] hose = new Hoses[500];                 // Hoses between components
Ports[] port = new Ports[1200];                 // Connection ports for all devices,

boolean released = false;                      // tests if mouse release for testing components
int connCount = 0;                             // Number of hose connections
int timerStart = 0;                            // used for timer clock for blinking unused ports
int interval = 0;
int redValue = 0;                              // allows cycling color of ports

PShape s;                                      // General image
PImage trash;                                  // port number is component # * 4 + 0 to 3
PImage toggleOn;                               // toggle switch declarations
PImage toggleOff;
int toggleCount;                               // Number of times turned "on"
String hoseMsg[][] = new String[1000][3];      // Holds messages for bad hose connections
int toggleX = 135;                             // position of toggle switch
int toggleY = 30;
boolean mouse= false;                          // track if mouse clicked
int [] teesChecked = new int [50];             // track which tees where checked while tracing paths
int teeCount=0;                                // number of tees places
int compCount=0;                                // number of component placed
int compLast=-1;                                // track last component selected
int trashX = 1050;                             // trash can position
int trashY = 80;

int numHoses = 0;                              // Total number of hoses, current hose drawing is numHoses
boolean routingHose = false;                   // Tracks if a hose is currently being routed
int x;
 // component constants for each ID
final int  tee = 0;
final int  res=1; 
final int  constDispPump = 2; 
final int  varDispPump = 3; 
final int  pressReg = 4; 
final int  openCenterValve = 5; 
final int  closedCenterValve = 6; 
final int  actuator = 7; 
final int  manualPump = 8; 
final int  accum = 9; 
final int  reliefValve = 10; 
final int  thermalReliefValve = 11; 
final int  pressGauge = 12; 
final int  shuttleValve = 13; 
final int  seqValve = 14; 
final int  priorityValve = 15; 
final int  checkValve = 16; 
final int  filter = 17; 
//18 – Fuse (might not use)  
final int  shutoffValve = 19; 
final int  Standpipe =20; 


void setup() {
  size(1200, 600);                            // display size
  surface.setResizable(true);
  background(255);                            // background color, 255 = white
  stroke(80);                                 // line thichness
                            
  trash = loadImage("trash.png");              // load image as defined for trsh can
  trash.resize(50,50); // optional              
  toggleOn = loadImage("RockerSW_Grn_1.gif");  // load switch images
  toggleOn.resize(30,50);
  toggleOff = loadImage("RockerSW_Red_0.gif");
  toggleOff.resize(30,50);
  createComponents();                          // Create all the objects for the screen
  }

 
void draw() {                 // redraws entire display each refresh cycle
   background(255);                            
   textSize(12);
   fill(128,0,0);
   textFont(createFont("Arial Bold",12));
   text("Hydraulics Sandbox\nVersion "+ ver +"\n" + date +"\nMartin Hebel (Code)\nKaren Johnson (SME)\nAviation Technologies\n" + 
        "Southern Illinois University",3,15);
 //  shape(s,400,350,40,40);  // rotation code - keep for now
 //  shape(s,500,300,80,80);
 //  shape(s,600,200,200,200);
 //  pushMatrix();
 //  translate(200,200);
 //  rotate(radians(50));
 //  shape(s,200,150,100,100);
 //  popMatrix();
    
   image(trash,trashX,trashY);                              // draw trash can
   image(toggleOn,toggleX,toggleY);                         // draw toggle on switch
   textSize(20);
   fill(0, 128, 0);
   text(toggleCount,155,100);                              // show number of times toggled on

   fill(0,0,0);
     
   if (!toggleFlag) image(toggleOff,toggleX,toggleY);  // if off, show off image
   boolean compFound= false;                          // used in tracking locating components and ports used
   boolean portFound1=false;                    
   boolean portFound= false;
 
  for (Components currentComponents : component) {    // when mouse moves, see if components of hose should be moved/made
    if (currentComponents.mouseOver() && (!routingHose) && (!compFound) && !currentComponents.checkConnected(port))
        { cursor(MOVE); 
          compFound = true; 
        }
  }
  
     // For each component, display what it needs to, 
     // check to see if mouse is dragging it             
  for (Components currentComponents : component) {
    currentComponents.display();          // display components
    currentComponents.mouseDragged();      // see if component being dragged
  
  }
  
  for (Ports currentPort : port) {  // used in routing hoses based on current location and status
    if (currentPort.mouseOver())       
        { cursor(CROSS); 
          portFound = true; 
          mouse = true;
          int hoseNum = currentPort.numHose;
          if (!toggleFlag || !terminal){
             if (hoseNum >= 0 && hoseMsg[hoseNum][0].length() > 0){                      // display message for hose connection
                  stroke(255,0,0);
                  fill(220);
                  rect(130, 0, 1000, 130, 7);
                  fill(0);
                  text(hoseMsg[hoseNum][0], 160,50);
                  text(hoseMsg[hoseNum][1], 160,75);
                  text(hoseMsg[hoseNum][2], 160,100);
             }
          }
        }
     if ((routingHose)) 
        { cursor(CROSS);  
          portFound1 = true;         // cross when selecting / routing
        }
        
     }

     if (!portFound && !compFound && !portFound1) // no longer routing a hose
            cursor(ARROW);


      
  redValue+=5; if (redValue > 255) redValue = 0;    // make open ports flash
  for (int c=0;c<285;c++){                          // display ports in use
    component[c].portDisplay(c, port, redValue);
  }
  
  if (routingHose){            // If routing a hose, display it
       hose[numHoses].finish(color(16,190,243));
       hose[numHoses].display();
    }
  
  for (int i=0; i<= numHoses;i++)     // display hoses in blue
    {
      if((toggleFlag) && (!routingHose) && (terminal))
      {
         hose[i].finish(color(0,0,255));
      }
      hose[i].display();
    }
}  // end draw()
 

 
void mousePressed() {
  boolean found= false;                   // loop each component until one found that is selected
 if (!routingHose)
   for (Components currentComponents : component) {
      if (!found) found = currentComponents.ComponentsMousePressed(port);
     }
  hoseCheck();                         // determine what action to take on hose 
  checkToggle();                        // check if toggle switch selected
 }

void mouseReleased() {
  released = true;
  if (timerStart == 0) timerStart = (int)(millis()/1000);
  for (Components currentComponents : component){         // find component mouse released on
    currentComponents.ComponentsMouseReleased(trash, trashX, trashY, port); 
  }
   if(toggleFlag) {             // if switch is off, check for proper hose connections.
       checkAll();
   }
} 


void hoseCheck(){
  if ((mouseButton==RIGHT) && routingHose){      // if routing, right mouse, delete (hide) hose
      hose[numHoses].hide();
      routingHose = false;
      port[hose[numHoses].getStartPortNum()].setHoseNum(-1);  // clear data and hose connections for it
      hose[numHoses].setStartComp(-1);                        // mark connections unused
      hose[numHoses].setStartPortNum(-1);
      hose[numHoses].setEndComp(-1);
      hose[numHoses].setEndPort(-1);
      hose[numHoses].setStartPortNum(-1);
      hose[numHoses].setEndPortNum(-1);
      numHoses++;
      return;
  }
  
  boolean found=false;
  for (Ports currentPort : port) {                  // Cycle though ports to find the one clicked
    int connNum = currentPort.connMousePressed();
    if (connNum != -1)                               // check in use, -1 not used
    if (!found){                                    // not found, keep checking
      if ((connNum != -1) && routingHose == true){  // if get back -1, means found unused, if routing, endhose 
        endHose(connNum);                            // Go end the hose connection
        found = true;                                // mark as found
      }
      else if ((connNum != -1) && routingHose == false){  // if found, and not routing, start a hose
        startHose(connNum);
        found=true;
      }  // end else
    } // end !found
  } // end for
  
  if (routingHose == true && found == false){        // if routing and not a connection, add more hose
        addHose();
  } // end hoseCheck
}

void startHose(int pNum){                                                   // start hose from port num 
     if (mouseButton == RIGHT){                                        // If right button, delete hose
       if (port[pNum].getHoseNum() != -1){
         hose[port[pNum].getHoseNum()].hide();
         int endNum;
         if (hose[port[pNum].getHoseNum()].getStartPortNum() == pNum)  // get other end of hose value
             endNum = hose[port[pNum].getHoseNum()].getEndPortNum();
         else
             endNum = hose[port[pNum].getHoseNum()].getStartPortNum();
         hose[port[pNum].getHoseNum()].setStartComp(-1);              // clear hose data
         hose[port[pNum].getHoseNum()].setStartPort(-1);
         hose[port[pNum].getHoseNum()].setEndComp(-1);
         hose[port[pNum].getHoseNum()].setEndPort(-1);
         hose[port[pNum].getHoseNum()].setStartPortNum(-1);
         hose[port[pNum].getHoseNum()].setEndPortNum(-1);
         port[pNum].setHoseNum(-1);                                  // clear both end ports
         port[endNum].setHoseNum(-1);
       } // end port check
       return;
     }// end delete
     
     if (port[pNum].getHoseNum() == -1){                                //if port clear, start the hose                               
       hose[numHoses].addHose(port[pNum].getX()+6, port[pNum].getY()+6,
                           port[pNum].getX()+6, port[pNum].getY()+6);
       hose[numHoses].setStartComp(port[pNum].getComp());               // set hose data 
       hose[numHoses].setStartPortNum(pNum);
       port[pNum].setHoseNum(numHoses);                                  // set port number for hose num
       routingHose = true;
     } // end if
} // end startHose

  
void addHose(){                                              // Add hose to mouse click
      hose[numHoses].addHose(0,0,mouseX, mouseY);
    }
 
void endHose(int pNum){                                                  // Ending a hose
      if (pNum == hose[numHoses].getStartPortNum()) return;              // same as start
      if (port[pNum].getHoseNum() == -1){                                // end port clear, the route hose
         hose[numHoses].addHose(port[pNum].getX(), port[pNum].getY(),
                             port[pNum].getX()+5, port[pNum].getY()+6);      
         hose[numHoses].addHose(port[pNum].getX(), port[pNum].getY(),    // once more to finish it
                             port[pNum].getX()+5, port[pNum].getY()+6);
         routingHose = false;
                                                                         // finish hose drawing
         hose[numHoses].setStartPort(hose[numHoses].getStartPortNum() % 4);  // store hose data
         hose[numHoses].setEndComp(port[pNum].getComp());
         hose[numHoses].setEndPort(pNum % 4);
         hose[numHoses].setEndPortNum(pNum);
         port[pNum].setHoseNum(numHoses);                                // store port data for hose
         hose[numHoses].finish(color(0,0,128));
         numHoses++;
      }
    }
 void createComponents(){              // create all components
    for(int x=0; x < 1000; x++){        // clear all hose message warnings
      hoseMsg[x][0] = "";
      hoseMsg[x][1] = "";
      hoseMsg[x][2] = "";
    }
  for (int i=0; i<5;i++)             // create 5 reservoirs, #1
    component[i] = new  Components(180,30,80,i,1, "Reservoir.png");

  for (int i=5; i<10;i++)             // create 5 pumps, #2
    component[i] = new  Components(260, 20,80,i,2, "Constant Displacement Pump.png");
  
  for (int i=10; i<15;i++)            // create 5 variable pumps, #3
    component[i] = new  Components(350,30,100,i,3, "Variable Displacement Pump.png");

  for (int i=15; i<25;i++)            // create 10 regulators, #4
    component[i] = new  Components(325+125, 30,80,i,4, "Pressure Regulator.png");
  
  for (int i=25; i<45;i++)            // create 20 open center valves, #5
    component[i] = new  Components(425+125, 20,60,i,5, "Open Center Selector Valve.png");

  for (int i=45; i<65;i++)           // create 20 closed center valves, #6
    component[i] = new  Components(425+125, 80,60,i,6, "Closed Center Selector Valve.png");
  
  for (int i=65; i<105;i++)           // create 40 Actuators, #7 
    component[i] = new  Components(500+125,30,100,i,7,"Actuator.png"); 
 
  for (int i=105; i<110;i++)           // create 5 manual pumps, #8 
    component[i] = new  Components(225+125,90,80,i,8,"Manual Pump.png"); 

  for (int i=110; i<115;i++)           // create 5 accumulators #9 
    component[i] = new  Components(560+125,70,60,i,9,"Accumulator.png"); 

  for (int i=115; i<125;i++)           // create 10 relief valves #10 
    component[i] = new  Components(500+125,70,50,i,10,"Relief Valve.png"); 
  
  for (int i=125; i<155;i++)           // create 20 pressure gauges #12 
    component[i] = new  Components(640+125,30,40,i,12,"Pressure Gauge.png");
  
  for (int i=155; i<165;i++)           // create 10 shuttle valves, #13
    component[i] = new  Components(700+125,30,80,i,13,"Shuttle Valve.png");  
  
  for (int i=165; i<175;i++)           // create 10 sequence valves, #14
    component[i] = new  Components(780+125,30,80,i,14,"Sequence Valve.png");    
  
  for (int i=175; i<185;i++)           // create 10 priority valves, #15
    component[i] = new  Components(850+125,30,100,i,15,"Priority Valve.png");
 
 // for (int i=185; i<200;i++)           // create 16 check valve - to right, #16
 //   component[i] = new  Components(700+125,100,60,i,16,"Check Valve.png");
    
  for (int i=185; i<210;i++)           // create 10 filters #17
    component[i] = new  Components(850+125,90,60,i,17,"Filter.png");
  
  for (int i=210; i<220;i++)           // create 10 shutoff valves #19
    component[i] = new  Components(145+125,100,60,i,19,"Shutoff Valve.png");

 // for (int i=220; i<235;i++)           // create 17 check valve - to left, #16
 //   component[i] = new  Components(780+125,100,60,i,20,"Check Valve (Return).png");
    
  for (int i=220; i<285;i++)           // create 50 Tees, #0 
    component[i] = new  Components(650+125,110,15,i,0,"Tees.jpg");
  
  for(int x=0;x<500;x++)              // create 500 hose lines
    hose[x]=new Hoses();
 
 for (int x=0; x<1200; x++){            // create 500 ports, 4 each comp
       port[x]=new Ports(0,-100,x, "Port.jpg");
       port[x].setHoseNum(-1);          // -1, mark unused
      }

} // end createComponents

void checkAll(){                  // checks legality of a hose connection
    for(int x=0; x < 1000; x++){
      hoseMsg[x][0] = "";
      hoseMsg[x][1] = "";
      hoseMsg[x][2] = "";
    }
  if(toggleFlag == false) return;                 // toggl off? Done 

  boolean finalResult = false;                   // start with false to indicate bad connection
  for (int i=0;(i < numHoses); i++)              // go through each hose used
     {
       connCount = 0;
       if (hose[i].visible()) {                  // if visible
          finalResult = checkHoses(i);                             // check for proper connections
          if (finalResult)  finalResult = test2Self(i);            // go ensure it doesn't connect to itself
          if (finalResult)  finalResult = checkBadTeeConnections(i);    // go run through not-allowed connection list
          if (finalResult)  finalResult = valves2Actuators(i);    // check valve/actuator agreement for both hoses
          setHose(i, finalResult);
         }
       }
 }

 void setHose(int i, boolean result){
         if (result){
          if (testHose(i,res,3))  // in return line, make dark green
            hose[i].finish(color(0,128,0));
          else
            hose[i].finish(color(0,255,0));    // good, normal green
        }
        else 
          hose[i].finish(color(253,0,0));      // bad, red
 }
 
boolean checkBadTeeConnections(int i){      // check if there is a T (counting filters and standpipe) //<>// //<>// //<>//
     if((hose[i].getEndComp() == tee  && hose[i].getStartComp() == tee) || 
           (hose[i].getEndComp() == tee && hose[i].getStartComp() == filter) || 
           (hose[i].getEndComp() == filter && hose[i].getStartComp() == tee) || 
           (hose[i].getEndComp() == filter && hose[i].getStartComp() == filter)||
           (hose[i].getEndComp() == Standpipe && hose[i].getStartComp() == tee) || 
           (hose[i].getEndComp() == Standpipe && hose[i].getStartComp() == filter)||
           (hose[i].getEndComp() == tee && hose[i].getStartComp() == Standpipe) || 
           (hose[i].getEndComp() == tee && hose[i].getStartComp() == Standpipe)||
           hose[i].getEndComp() == tee ||
           hose[i].getStartComp() == tee)
           {
            if (testBad2TeesList(i)) return false;     // check Do Not Connect list with T's
           }
      // ensure left check valve is only in return line   
      if ((hose[i].getStartComp() == Standpipe || hose[i].getEndComp() == Standpipe) && !testHose(i,res,varDispPump))
          return false;
      return true;
}

boolean checkHoses(int i) {
    boolean result = false;   
    result = testHosePart(i);
    return result;
  }

        // proper connections, return true when found one
boolean testHosePart(int i){
        // first, non-legal connections, with warning messages
   if (testHose(i,res,3,constDispPump,0)){                           
       hoseMsg[i][0]="This does not provide a positive supply of fluid to the powered pump.";
       return false;
   }
   
   if (testHose(i,manualPump,1,constDispPump,0)){                        
      hoseMsg[i][0]="This incorrectly routes pressurized fluid to the supply line of the powered pump.";
      return false;
   }
   
   if (testHose(i,manualPump,0,constDispPump,1)){                       
    hoseMsg[i][0]="This incorrectly routes pressurized fluid to the supply line of the manual pump.";
    return false;
   }
    
   if (testHose(i,manualPump,1,varDispPump,1)) {                       
    hoseMsg[i][0]="This incorrectly routes pressurized fluid to the supply line of the powered pump.";
    return false;
   }
   
   if (testHose(i,manualPump,0,varDispPump,0)){ 
      hoseMsg[i][0]="This incorrectly routes pressurized fluid to the supply line of the manual pump.";
     return false;
   }
   
   if (testHose(i,res,0,shutoffValve,1)) {                           
    hoseMsg[i][0]="This installs the shutoff valve backwards.";
    return false;
   }
   
  if (testHose(i,res,1,shutoffValve,1)){
    hoseMsg[i][0]="This installs the shutoff valve backwards.";
    return false;
   }
   
   if (testHose(i,res,0,closedCenterValve,0)){
    hoseMsg[i][0]="There needs to a pump to pressurize the fluid from the reservoir";
    hoseMsg[i][1]="before routing to the selector valve";
    return false;
   }
        
   if (testHose(i,res,1,closedCenterValve,0)){
      hoseMsg[i][0]="There needs to a pump to pressurize the fluid from the reservoir";
      hoseMsg[i][1]="before routing to the selector valve";
      return false;
   }
        
   if (testHose(i,res,0,openCenterValve,0)){
      hoseMsg[i][0]="There needs to a pump to pressurize the fluid from the reservoir";
      hoseMsg[i][1]="before routing to the selector valve";
      return false;
   }
   
   if (testHose(i,res,1,openCenterValve,0)){
      hoseMsg[i][0]="There needs to a pump to pressurize the fluid from the reservoir";
      hoseMsg[i][1]="before routing to the selector valve";
      return false;
   }
        if (testHose(i,res,0,tee,0))                                       return false; 
        if (testHose(i,res,1,tee,0))                                       return false; // When Tee is used for manual and powered pump
//  This prevents the fluid below the standpipe in the reservoir from supplying the manual pump
        if (testHose(i,res,0,tee,1))                                       return false;
        if (testHose(i,res,1,tee,1))                                       return false; 
        if (testHose(i,res,0,tee,2))                                       return false;
        if (testHose(i,res,1,tee,2))                                       return false;
   
   if (testHose(i,openCenterValve,0,openCenterValve,0)){
      hoseMsg[i][0]="This bypasses the first selector valve in the series";
       return false;
   }
    
   if (testHose(i,openCenterValve,1,openCenterValve,1)){
      hoseMsg[i][0]="This bypasses the subsequent valve in the series";
      // hoseMsg[i][1]="line 2";
      return false;
   }
   if (testCompHose(i,res,actuator)){ 
      hoseMsg[i][0]="This connection would not supply pressure to the actuator"; 
      return false; 
} 

if (testCompHose(i,res,accum)){ 
      hoseMsg[i][0]="An accumulator must be in a pressurized line"; 
      return false; 
} 

if (testCompHose(i,res,shuttleValve)){ 
      hoseMsg[i][0]="Shuttle valves are used with actuators"; 
      return false; 
} 

if (testCompHose(i,res,actuator)){ 
      hoseMsg[i][0]="This connection would not supply pressure to the actuator"; 
      return false; 
} 

if (testCompHose(i,res,seqValve)){ 
      hoseMsg[i][0]="Sequence valves are used to sequence actuators"; 
      return false; 
} 

if (testCompHose(i,res,priorityValve)){ 
      hoseMsg[i][0]="Priority valves are used to sequence actuators"; 
      return false; 
} 

if (testCompHose(i,constDispPump,shuttleValve)){ 
      hoseMsg[i][0]="Shuttle valves are used for emergency sources of pressure"; 
      return false; 
} 

if (testCompHose(i,constDispPump,priorityValve)){ 
     hoseMsg[i][0]="Priority valves are used to sequence actuators"; 
     return false; 
} 

if (testCompHose(i,constDispPump,seqValve)){ 
     hoseMsg[i][0]="Sequence valves are used to sequence actuators"; 
     return false; 
} 

if (testCompHose(i,varDispPump,shuttleValve)){ 
     hoseMsg[i][0]="Shuttle valves are used for emergency sources of pressure"; 
     return false; 
} 

if (testCompHose(i,varDispPump,priorityValve)){ 
     hoseMsg[i][0]="Priority valves are used to sequence actuators"; 
     return false; 
} 

if (testCompHose(i,varDispPump,seqValve)){ 
     hoseMsg[i][0]="Sequence valves are used to sequence actuators"; 
     return false; 
} 

if (testCompHose(i,priorityValve,seqValve)){ 
     hoseMsg[i][0]="This connection is duplicative"; 
     return false; 
} 

   // legal connections
        if (testHose(i,res,0,constDispPump,0) && !testIfConnection(res,1,manualPump,0)) return true; 
        if (testHose(i,res,0,varDispPump,1) && !testIfConnection(res,1,manualPump,0))   return true;
        if (testHose(i,res,1,constDispPump,0))                                          return true;
        if (testHose(i,res,1,varDispPump,1))                                            return true;  
        if (testHose(i,pressReg,2,closedCenterValve,0))                                 return true; 
        if (testHose(i,pressReg,2,closedCenterValve,1))                                 return true;
        if (testHose(i,constDispPump,1,pressReg,1))                                     return true;   
        if (testHose(i,openCenterValve,1,openCenterValve,0))                            return true; 
        if (testHose(i,pressReg,0,res,3))                                               return true;  
        if (testHose(i,constDispPump,1,openCenterValve,0))                              return true;  
        if (testHose(i,varDispPump,0,closedCenterValve,0))                              return true;
        if (testHose(i,openCenterValve,0,res,3))                                        return true;   
        if (testHose(i,openCenterValve,1,res,3))                                        return true;  
        if (testHose(i,closedCenterValve,0,res,3))                                      return true;   
        if (testHose(i,closedCenterValve,1,res,3))                                      return true;  
        if (testTee(i))                                                                 return true; // Tee to Tee
        if (testHose(i,res,0,manualPump,0))                                             return true;     
        if (testHose(i,manualPump,1,constDispPump,1) && testIfConnection(manualPump,1,pressReg,1))        return true;   
        if (testHose(i,manualPump,1,constDispPump,1) && testIfConnection(manualPump,1,openCenterValve,0)) return true;
        if (testHose(i,manualPump,1,pressReg,2) && testIfConnection(manualPump,1,closedCenterValve,0))    return true;   
        if (testHose(i,manualPump,1,closedCenterValve,0) && testIfConnection(manualPump,1,pressReg,2))    return true; 
        if (testHose(i,manualPump,1,varDispPump,0) && testIfConnection(manualPump,1,closedCenterValve,0))    return true;  
        if (testHose(i,accum,0,pressReg,2) && testIfConnection(accum,0,closedCenterValve,0))                 return true;   
        if (testHose(i,accum,0,constDispPump,1) && testIfConnection(accum,0,openCenterValve,0))              return true;    
        if (testHose(i,accum,0,varDispPump,0) && testIfConnection(accum,0,closedCenterValve,0))              return true;    
        if (testHose(i,reliefValve,1,res,3))                                                                 return true;         
        if (testHose(i,reliefValve,0,pressReg,2) && testIfConnection(reliefValve,0,closedCenterValve,0))     return true;    
        if (testHose(i,reliefValve,0,constDispPump,1) && testIfConnection(reliefValve,0,openCenterValve,0))  return true;     
        if (testHose(i,reliefValve,0,varDispPump,0) && testIfConnection(reliefValve,0,closedCenterValve,0))  return true;     
        if (testHose(i,pressGauge,0,varDispPump,0) && testIfConnection(pressGauge,0,closedCenterValve,0))    return true;    
        if (testHose(i,pressGauge,0,pressReg,2) && testIfConnection(pressGauge,0,closedCenterValve,0))       return true;   
        if (testHose(i,pressGauge,0,constDispPump,1) && testIfConnection(pressGauge,0,openCenterValve,0))    return true;   
    
        if (testHose(i,openCenterValve,2,actuator,0))   return true;   
        if (testHose(i,openCenterValve,3,actuator,0))   return true;   
        if (testHose(i,openCenterValve,2,actuator,1))   return true;   
        if (testHose(i,openCenterValve,3,actuator,1))   return true;   

        if (testHose(i,closedCenterValve,2,actuator,0)) return true;         
        if (testHose(i,closedCenterValve,3,actuator,0)) return true;         
        if (testHose(i,closedCenterValve,2,actuator,1)) return true;         
        if (testHose(i,closedCenterValve,3,actuator,1)) return true;         

        if (testHose(i,openCenterValve,2,shuttleValve,0) && testIfConnection(actuator,1,shuttleValve,2)) return true;   // 13 MUST connect to 5 AND 7   **
        if (testHose(i,actuator,1,shuttleValve,2) && testIfConnection(openCenterValve,2,shuttleValve,0)) return true;
        if (testHose(i,openCenterValve,3,shuttleValve,0) && testIfConnection(actuator,1,shuttleValve,2)) return true;   
        if (testHose(i,actuator,1,shuttleValve,2) && testIfConnection(openCenterValve,3,shuttleValve,0)) return true;
        if (testHose(i,openCenterValve,2,shuttleValve,0) && testIfConnection(actuator,0,shuttleValve,2)) return true;   
        if (testHose(i,actuator,0,shuttleValve,2) && testIfConnection(openCenterValve,2,shuttleValve,0)) return true;
        if (testHose(i,openCenterValve,3,shuttleValve,0) && testIfConnection(actuator,0,shuttleValve,2)) return true;   
        if (testHose(i,actuator,0,shuttleValve,2) && testIfConnection(openCenterValve,3,shuttleValve,0)) return true;
 
        if (testHose(i,closedCenterValve,2,shuttleValve,0) && testIfConnection(actuator,1,shuttleValve,2)) return true;   // 13 MUST connect to 6 AND 7   **
        if (testHose(i,actuator,1,shuttleValve,2) && testIfConnection(closedCenterValve,2,shuttleValve,0)) return true;
        if (testHose(i,closedCenterValve,3,shuttleValve,0) && testIfConnection(actuator,1,shuttleValve,2)) return true;   
        if (testHose(i,actuator,1,shuttleValve,2) && testIfConnection(closedCenterValve,3,shuttleValve,0)) return true;
        if (testHose(i,closedCenterValve,2,shuttleValve,0) && testIfConnection(actuator,0,shuttleValve,2)) return true;   
        if (testHose(i,actuator,0,shuttleValve,2) && testIfConnection(closedCenterValve,2,shuttleValve,0)) return true;
        if (testHose(i,closedCenterValve,3,shuttleValve,0) && testIfConnection(actuator,0,shuttleValve,2)) return true;   
        if (testHose(i,actuator,0,shuttleValve,2) && testIfConnection(closedCenterValve,3,shuttleValve,0)) return true;
       
        if (testHose(i,0,1,seqValve,0) && testIfConnection(seqValve,1,actuator,0))               return true;
        if (testHose(i,0,2,seqValve,0) && testIfConnection(seqValve,1,actuator,0))               return true;   
        if (testHose(i,0,3,seqValve,0) && testIfConnection(seqValve,1,actuator,0))               return true;   
        if (testHose(i,0,pressReg,seqValve,0) && testIfConnection(seqValve,1,actuator,0))        return true;   
        
        if (testHose(i,openCenterValve,2,seqValve,0) && testIfConnection(seqValve,1,actuator,0)) return true;   
        if (testHose(i,openCenterValve,3,seqValve,0) && testIfConnection(seqValve,1,actuator,0)) return true;   
        if (testHose(i,openCenterValve,3,seqValve,0) && testIfConnection(seqValve,1,priorityValve,0)) return true;
        
        if (testHose(i,constDispPump,1,filter,0)) return true;   
        
        if (testHose(i,seqValve,1,actuator,0))     return true;  
        if (testHose(i,seqValve,1,actuator,1))     return true; 
        if (testHose(i,seqValve,1,priorityValve,0))     return true; 
           
        if (testHose(i,closedCenterValve,2,priorityValve,0))    return true;   
        if (testHose(i,closedCenterValve,3,priorityValve,0))    return true;   
        if (testHose(i,openCenterValve,2,priorityValve,0))      return true;   
        if (testHose(i,openCenterValve,3,priorityValve,0))      return true;   
        if (testHose(i,priorityValve,1,actuator,1))             return true;  
        if (testHose(i,priorityValve,0,actuator,1))             return true; 
        if (testHose(i,priorityValve,1,actuator,0))             return true; 
 
        if (testHose(i,filter,0,constDispPump,1))       return true;  
        if (testHose(i,filter,1,openCenterValve,0))     return true;
 
        if (testHose(i,filter,0,pressReg,2))            return true;  
        if (testHose(i,filter,1,closedCenterValve,0))   return true;
      
        if (testHose(i,filter,0,varDispPump,0))         return true;   
        if (testHose(i,filter,1,closedCenterValve,0))   return true;
      
        if (testHose(i,filter,1,openCenterValve,1))     return true;  
        if (testHose(i,filter,0,res,3))                 return true;
      
        if (testHose(i,filter,1,closedCenterValve,1))   return true;   
        if (testHose(i,filter,0,res,3))                 return true;
              
        if (testHose(i,res,0,shutoffValve,0) && testIfConnection(shutoffValve,1,constDispPump,0))   return true;   
        if (testHose(i,shutoffValve,1,constDispPump,0) && testIfConnection(shutoffValve,0,1,0))     return true;

        if (testHose(i,res,0,shutoffValve,0) && testIfConnection(shutoffValve,1,varDispPump,1))     return true;   
        if (testHose(i,shutoffValve,1,varDispPump,1) && testIfConnection(shutoffValve,0,1,0))       return true;
      
        if (testHose(i,res,1,shutoffValve,0) && testIfConnection(shutoffValve,1,constDispPump,0))   return true; 
        if (testHose(i,shutoffValve,1,constDispPump,0) && testIfConnection(shutoffValve,0,1,1))     return true;

        if (testHose(i,res,1,shutoffValve,0) && testIfConnection(shutoffValve,1,varDispPump,1))     return true; 
        if (testHose(i,shutoffValve,1,varDispPump,1) && testIfConnection(shutoffValve,0,res,1))     return true;

        if (testHose(i,res,1,shutoffValve,0) && testIfConnection(shutoffValve,1,checkValve,0))      return true; 
        if (testHose(i,shutoffValve,1,checkValve,0) && testIfConnection(shutoffValve,0,res,1))      return true;

        if (testHose(i,res,0,shutoffValve,0) && testIfConnection(shutoffValve,1,checkValve,0))      return true; 
        if (testHose(i,shutoffValve,1,checkValve,0) && testIfConnection(shutoffValve,0,res,0))      return true;
        
        if (testHose(i,res,0,shutoffValve,0) && testIfConnection(shutoffValve,1,manualPump,0))      return true; 
        if (testHose(i,shutoffValve,1,manualPump,0) && testIfConnection(shutoffValve,0,1,0))        return true;
        
        if (testCompHose(i,res,actuator)){  //<>//
             hoseMsg[i][0]="Don’t do that!"; 
             // hoseMsg[i][0]="line 2 if needed";
             return false; 
           } 
           
        if (testCompHose(i,openCenterValve,closedCenterValve)){  //<>//
             hoseMsg[i][0]="Don’t do that!"; 
             // hoseMsg[i][0]="line 2 if needed";
             return false; 
           }
        
        return false;   // if none of the above, false

}
        // these should NOT be connected together through a Tee
  boolean testBad2TeesList(int i){
      if (test2TeesComp(i,res,actuator)){
      hoseMsg[i][0] ="This prevents an actuator from being actuated in both directions.";
      return true;
  }
  
      if (test2TeesComp(i,res,accum)){
        hoseMsg[i][0] ="This incorrectly routes supply fluid to the accumulator.";
        return true;
  }
      if (test2TeesComp(i,res,pressGauge)){
        hoseMsg[i][0] ="This incorrectly routes supply fluid to the pressure gauge.";
        return true;
      }
      
      if (test2TeesComp(i,res,shuttleValve)){
        hoseMsg[i][0] ="This incorrectly routes an alternate source or pressure to the reservoir.";
        return true;
      }
      
      if (test2TeesComp(i,res,seqValve))                 return true;
      if (test2TeesComp(i,res,priorityValve))            return true;

      if (test2TeesComp(i,constDispPump,varDispPump)){
        hoseMsg[i][0] ="These two types of pumps cannot be used together in the same system.";
        return true;
  }
      if (test2TeesComp(i,constDispPump,actuator)){
        hoseMsg[i][0] ="This incorrectly routes pressurized fluid to only one side of the actuator."; //<>//
        return true;
  }

      if (test2TeesComp(i,constDispPump,shuttleValve))  return true;
      if (test2TeesComp(i,constDispPump,seqValve))       return true; 
      if (test2TeesComp(i,constDispPump,priorityValve))  return true;
      
      if (test2TeesComp(i,varDispPump,pressReg)) {
        hoseMsg[i][0] ="A variable displacement pump internally regulates system pressure"; 
        return true;
  }
      if (test2TeesComp(i,varDispPump,openCenterValve)) {
        hoseMsg[i][0] ="Variable displacement pumps do not require open center selector valves";
        return true;
  }
  
      if (test2TeesComp(i,varDispPump,actuator)) {
        hoseMsg[i][0] ="This incorrectly routes pressurized fluid to only one side of the actuator.";
        return true;
  }

      if (test2TeesComp(i,varDispPump,shuttleValve))     return true;
      if (test2TeesComp(i,varDispPump,seqValve))         return true;
      if (test2TeesComp(i,varDispPump,priorityValve))    return true;
      
      if (test2TeesComp(i,manualPump,actuator)) {
        hoseMsg[i][0] ="This incorrectly routes pressurized fluid to only one side of the actuator.";
        return true;
  }

      if (test2TeesComp(i,manualPump,shuttleValve))      return true;
      if (test2TeesComp(i,manualPump,seqValve))          return true;
      if (test2TeesComp(i,manualPump,priorityValve))     return true;

      if (test2TeesComp(i,pressReg,actuator)){
        hoseMsg[i][0] ="This incorrectly allows the pressure regulator to protect only one portion of the system.";
        return true;
  }

      if (test2TeesComp(i,pressReg,shuttleValve))        return true;
      if (test2TeesComp(i,pressReg,seqValve))            return true;  
      if (test2TeesComp(i,pressReg,priorityValve))       return true;
      if (test2TeesComp(i,openCenterValve,closedCenterValve) && !testHose(i,res,3))  return true; 
      
      if (test2TeesComp(i,actuator,accum)){
        hoseMsg[i][0] ="This incorrectly allows the accumulator to support only one portion of the system.";
        return true;
    }
      
      if (test2TeesComp(i,actuator,reliefValve)){
          hoseMsg[i][0] ="This incorrectly allows the relief valve to protect only one portion of the system.";
          return true; //<>//
      }

      if (test2TeesComp(i,actuator,pressGauge)){
        hoseMsg[i][0] ="This incorrectly allows pressure from only one portion of the system to be indicated.";
        return true;
  }

      if (test2TeesComp(i,actuator,filter)){
        hoseMsg[i][0] ="Filters are found in common lines.";
        return true;
  }
      if (test2TeesComp(i,accum,shuttleValve))           return true;
      if (test2TeesComp(i,accum,seqValve))               return true;
      if (test2TeesComp(i,accum,priorityValve))          return true;
      if (test2TeesComp(i,reliefValve,shuttleValve))     return true;
      if (test2TeesComp(i,reliefValve,seqValve))         return true;
      if (test2TeesComp(i,reliefValve,priorityValve))    return true;
      if (test2TeesComp(i,pressGauge,shuttleValve))      return true;
      if (test2TeesComp(i,pressGauge,seqValve))          return true;
      if (test2TeesComp(i,pressGauge,priorityValve))     return true;
      if (test2TeesComp(i,shuttleValve,priorityValve))   return true;
      if (test2TeesComp(i,shuttleValve,filter))          return true;
      
      if (test2TeesComp(i,seqValve,priorityValve)){
       hoseMsg[i][0] ="A sequence and priority valve must be on separate sides of an actuator.";
    return true;
  }
      if (test2TeesComp(i,seqValve,filter))              return true;  
      if (test2TeesComp(i,priorityValve,filter))         return true;
      
      return false;
}

boolean testCompHose(int hoseNum, int comp1, int comp2){  // test for any connections between components without port number
  return testHose(hoseNum, comp1, -1,comp2, -1);
}

        // test connection between components based on port numbers
boolean testHose(int hoseNum,int comp1, int port1, int comp2, int port2)
{    // test headers
      resetTeeCount();
     int compNum=0;
     connCount = 0;
     println(hose[hoseNum].getStartComp());
      //     find out if it's a tee, filter or standPipe. If so, get the component number instead of type
     if (hose[hoseNum].getStartComp() == tee) 
       compNum = hose[hoseNum].getStartCompNum();
     if (hose[hoseNum].getEndComp() == tee) 
       compNum = hose[hoseNum].getEndCompNum();
     if (hose[hoseNum].getStartComp() == filter) 
       compNum = hose[hoseNum].getStartCompNum();
     if (hose[hoseNum].getEndComp() == filter) 
       compNum = hose[hoseNum].getEndCompNum();
     if (hose[hoseNum].getStartComp() == Standpipe)   
       compNum = hose[hoseNum].getStartCompNum();
     if (hose[hoseNum].getEndComp() == Standpipe) 
       compNum = hose[hoseNum].getEndCompNum();
     
     // if there is a tee, filter or standpipe, run checks
     if ((hose[hoseNum].getEndComp() == tee || hose[hoseNum].getStartComp() == tee) || 
         (hose[hoseNum].getEndComp() == filter || hose[hoseNum].getStartComp() == filter) ||
         (hose[hoseNum].getEndComp() == tee || hose[hoseNum].getStartComp() == filter) ||
         (hose[hoseNum].getEndComp() == filter || hose[hoseNum].getStartComp() == tee) ||
         (hose[hoseNum].getEndComp() == tee || hose[hoseNum].getStartComp() == Standpipe) ||
         (hose[hoseNum].getEndComp() == Standpipe || hose[hoseNum].getStartComp() == tee) ||
         (hose[hoseNum].getEndComp() == Standpipe || hose[hoseNum].getStartComp() == filter) ||
         (hose[hoseNum].getEndComp() == filter || hose[hoseNum].getStartComp() == Standpipe)||
         (hose[hoseNum].getEndComp() == tee || 
         hose[hoseNum].getStartComp() == tee))
       {
         // Go trace through Tees
         boolean sameType = false;
         if (comp1 == comp2) sameType = true;
       if(hose[hoseNum].getStartComp() == comp1 && (hose[hoseNum].getStartPort() == port1 || port1 == -1)){
          if (checkTeeHoses(compNum, comp2, port2, sameType)) return true;
       }
       if(hose[hoseNum].getEndComp() == comp1 && (hose[hoseNum].getEndPort() == port1 || port1 == -1)){
             if (checkTeeHoses(compNum, comp2, port2, sameType)) return true;
       }
       if(hose[hoseNum].getStartComp() == comp2 && (hose[hoseNum].getStartPort() == port2 || port2 == -1)){
          if (checkTeeHoses(compNum, comp1, port1, sameType)) return true;
       }
       if(hose[hoseNum].getEndComp() == comp2 && (hose[hoseNum].getEndPort() == port2 || port2 == -1)){
             if (checkTeeHoses(compNum, comp1, port1, sameType)) return true;
       }
     }
     else
       // No Tees, etc, check connection
     {  if ((hose[hoseNum].getStartComp() == comp1 && port1 == -1 &&  // check forward and backward
         hose[hoseNum].getEndComp() == comp2 && port2 == -1) ||
         (hose[hoseNum].getStartComp() == comp2 && port1 == -1 &&
          hose[hoseNum].getEndComp() == comp1 && port2 == -1))
       return true;
       
       if ((hose[hoseNum].getStartComp() == comp1 && hose[hoseNum].getStartPort() == port1 &&  // check forward and backward
         hose[hoseNum].getEndComp() == comp2 && hose[hoseNum].getEndPort() == port2) ||
         (hose[hoseNum].getStartComp() == comp2 && hose[hoseNum].getStartPort() == port2 &&
          hose[hoseNum].getEndComp() == comp1 && hose[hoseNum].getEndPort() == port1))
        return true;
      else    
        return false;
     }
     return false;
}

boolean testHoseNum(int hoseNum,int comp1, int port1, int comp2, int port2)
{    // test headers
      resetTeeCount();
     int compNum=0;
      //     find out if it's a tee, filter or standPipe. If so, get the component number instead of type
     if (hose[hoseNum].getStartComp() == tee) compNum = hose[hoseNum].getStartCompNum();
     if (hose[hoseNum].getEndComp() == tee) compNum = hose[hoseNum].getEndCompNum();
     if (hose[hoseNum].getEndComp() == filter) compNum = hose[hoseNum].getEndCompNum();
     if (hose[hoseNum].getEndComp() == filter) compNum = hose[hoseNum].getEndCompNum();
     if (hose[hoseNum].getStartComp() == Standpipe) compNum = hose[hoseNum].getStartCompNum();
     if (hose[hoseNum].getEndComp() == Standpipe) compNum = hose[hoseNum].getEndCompNum();
     
     // if there is a tee, filter or standpipe, run checks
     if ((hose[hoseNum].getEndComp() == tee || hose[hoseNum].getStartComp() == tee) || 
         (hose[hoseNum].getEndComp() == filter || hose[hoseNum].getStartComp() == filter) ||
         (hose[hoseNum].getEndComp() == tee || hose[hoseNum].getStartComp() == filter) ||
         (hose[hoseNum].getEndComp() == filter || hose[hoseNum].getStartComp() == tee) ||
         (hose[hoseNum].getEndComp() == tee || hose[hoseNum].getStartComp() == Standpipe) ||
         (hose[hoseNum].getEndComp() == Standpipe || hose[hoseNum].getStartComp() == tee) ||
         (hose[hoseNum].getEndComp() == Standpipe || hose[hoseNum].getStartComp() == filter) ||
         (hose[hoseNum].getEndComp() == filter || hose[hoseNum].getStartComp() == Standpipe)||
         (hose[hoseNum].getEndComp() == tee || 
         hose[hoseNum].getStartComp() == tee))
       {
         boolean sameType = false;
         if (comp1 == comp2) sameType = true;
         // Go trace through Tees
       if(hose[hoseNum].getStartCompNum() == comp1 && hose[hoseNum].getStartPort() == port1){
          if (checkTeeCompNum(compNum, comp2, port2,sameType)) return true;
       }
       if(hose[hoseNum].getEndCompNum() == comp1 && hose[hoseNum].getEndPort() == port1){
             if (checkTeeCompNum(compNum, comp2, port2,sameType)) return true;
       }
       if(hose[hoseNum].getStartCompNum() == comp2 && hose[hoseNum].getStartPort() == port2){
          if (checkTeeCompNum(compNum, comp1, port1, sameType)) return true;
       }
       if(hose[hoseNum].getEndCompNum() == comp2 && hose[hoseNum].getEndPort() == port2){
             if (checkTeeCompNum(compNum, comp1, port1, sameType)) return true;
       }
     }
     else
       // No Tees, etc, check connection
     {
       if ((hose[hoseNum].getStartCompNum() == comp1 && hose[hoseNum].getStartPort() == port1 &&  // check forward and backward
         hose[hoseNum].getEndCompNum() == comp2 && hose[hoseNum].getEndPort() == port2) ||
         (hose[hoseNum].getStartCompNum() == comp2 && hose[hoseNum].getStartPort() == port2 &&
          hose[hoseNum].getEndCompNum() == comp1 && hose[hoseNum].getEndPort() == port1))
        return true;
      else    
        return false;
     }
     return false;
}


boolean testHose(int hoseNum,int comp1, int port1)
{    // test headers
      resetTeeCount();
     int compNum=0;
     connCount = 0;
     if (hose[hoseNum].getStartComp() == tee) compNum = hose[hoseNum].getStartCompNum();
     if (hose[hoseNum].getEndComp() == tee) compNum = hose[hoseNum].getEndCompNum();
     if (hose[hoseNum].getStartComp() == filter) compNum = hose[hoseNum].getEndCompNum();
     if (hose[hoseNum].getEndComp() == filter) compNum = hose[hoseNum].getEndCompNum();
     if (hose[hoseNum].getStartComp() == Standpipe) compNum = hose[hoseNum].getStartCompNum();
     if (hose[hoseNum].getEndComp() == Standpipe) compNum = hose[hoseNum].getEndCompNum();
     
     if ((hose[hoseNum].getEndComp() == tee || hose[hoseNum].getStartComp() == tee) || 
         (hose[hoseNum].getEndComp() == filter || hose[hoseNum].getStartComp() == filter) ||
         (hose[hoseNum].getEndComp() == Standpipe || hose[hoseNum].getStartComp() == Standpipe))
       {
          if (checkTeeHoses(compNum, comp1, port1, false)) return true;
       }
     else
     {
       if (hose[hoseNum].getStartComp() == comp1 && hose[hoseNum].getStartPort() == port1) return true; // && hose[hoseNum].getStartPort() == port1 &&  // check forward and backward
       if (hose[hoseNum].getEndComp() == comp1 && hose[hoseNum].getEndPort() == port1) return true; // && hose[hoseNum].getEndPort() == port2) ||
     }
     return false;
}

boolean test2TeesComp(int i,int comp1, int comp2)
{    // test headers
      resetTeeCount();
      compCount=0;
      compLast=-1;
      int compTee=0;
     
     if ((hose[i].getEndComp() == tee && hose[i].getStartComp() == tee) || 
         (hose[i].getEndComp() == tee && hose[i].getStartComp() == filter) || 
         (hose[i].getEndComp() == filter && hose[i].getStartComp() == tee) || 
         (hose[i].getEndComp() == filter && hose[i].getStartComp() == filter)|| 
         (hose[i].getEndComp() == tee && hose[i].getStartComp() == Standpipe)|| 
         (hose[i].getEndComp() == Standpipe && hose[i].getStartComp() == tee)|| 
         (hose[i].getEndComp() == filter && hose[i].getStartComp() == Standpipe)|| 
         (hose[i].getEndComp() == Standpipe && hose[i].getStartComp() == filter) ||
         (hose[i].getEndComp() == tee || 
         hose[i].getStartComp() == tee)){
           if (hose[i].getEndComp() == tee) {
               compTee = hose[i].getEndPortNum()/4;
           }
           if (hose[i].getEndComp() == filter) {
               compTee = hose[i].getEndPortNum()/4;
           }
           if (hose[i].getEndComp() == Standpipe) {
               compTee = hose[i].getEndPortNum()/4;
           }
           if (hose[i].getStartComp() == tee) {
               compTee = hose[i].getStartPortNum()/4;
           }
           if (hose[i].getStartComp() == filter) {
               compTee = hose[i].getStartPortNum()/4;
           }
           if (hose[i].getStartComp() == Standpipe) {
               compTee = hose[i].getStartPortNum()/4;
           }
          
           if (compTee > 0)
              if (checkTeeComp(compTee, comp1, comp2)) return true;
         }
          return false;
}

boolean test2Self(int i)
{    // test headers

      resetTeeCount();
      compCount=0;
      int compTee = 0;
      int startComp=1;
      
      if ((hose[i].getEndComp() == tee && hose[i].getStartComp() == tee) || 
         (hose[i].getEndComp() == tee && hose[i].getStartComp() == filter) || 
         (hose[i].getEndComp() == filter && hose[i].getStartComp() == tee) || 
         (hose[i].getEndComp() == filter && hose[i].getStartComp() == filter)|| 
         (hose[i].getEndComp() == tee && hose[i].getStartComp() == Standpipe)|| 
         (hose[i].getEndComp() == Standpipe && hose[i].getStartComp() == tee)|| 
         (hose[i].getEndComp() == filter && hose[i].getStartComp() == Standpipe)|| 
         (hose[i].getEndComp() == Standpipe && hose[i].getStartComp() == filter)
         ){
           return true;
       }
           if (hose[i].getEndComp() == tee) {
               compTee = hose[i].getEndPortNum()/4;
               startComp = hose[i].getStartCompNum();
           }
           if (hose[i].getEndComp() == filter) {
               compTee = hose[i].getEndPortNum()/4;
               startComp = hose[i].getStartCompNum();
           }
           if (hose[i].getEndComp() == Standpipe) {
               compTee = hose[i].getEndPortNum()/4;
               startComp = hose[i].getStartCompNum();
           }
           if (hose[i].getStartComp() == tee) {
               compTee = hose[i].getStartPortNum()/4;
               startComp = hose[i].getEndCompNum();
           }
           if (hose[i].getStartComp() == filter) {
               compTee = hose[i].getStartPortNum()/4;
               startComp = hose[i].getEndCompNum();
           }
           if (hose[i].getStartComp() == Standpipe) {
               compTee = hose[i].getStartPortNum()/4;
               startComp = hose[i].getEndCompNum();
           }
          
           if (compTee > 0)
              if (checkTeeCompNum(compTee, startComp, startComp, false)) return false;
          return true;
}

boolean valves2Actuators(int i){    
  // verify from the selector valves, though components, back is to the same selector valve.
      compCount=0;
      int compTee = 0;
      int startComp=0;
      int endComp=0;
      int startPort=0;
      int endPort=0;
      boolean result = false;
      
      // not comp of interest
      startComp = hose[i].getStartComp();
      endComp = hose[i].getEndComp();
      startPort = hose[i].getStartPort();
      endPort = hose[i].getEndPort();
      
      if (startComp != openCenterValve && startComp != closedCenterValve 
          && endComp != openCenterValve && endComp != closedCenterValve)
             return true;
             
      // not port of interest on select valves
      if (startComp == openCenterValve && (startPort != 2 && startPort != 3))
          return true;

      if (endComp == openCenterValve && (endPort != 2 && endPort != 3))
          return true;
          
      if (startComp == closedCenterValve && startPort != 2 && startPort != 3)
          return true;
          
      if (endComp == closedCenterValve && endPort != 2 && endPort != 3)
          return true;

      
      resetTeeCount();
      // connections between start and finish
      if (startComp == tee && endComp == tee)                   return true;
      if (startComp == tee && endComp == seqValve)            return true;
      if (startComp == tee && endComp == priorityValve)       return true;
      if (startComp == tee && endComp == seqValve)            return true;
      if (startComp == tee && endComp == actuator)            return true;
      if (startComp == tee && endComp == shuttleValve)        return true;
      
      if (startComp == tee && endComp == tee)                   return true;
      if (endComp == 0 && startComp == seqValve)            return true;
      if (endComp == 0 && startComp == priorityValve)       return true;
      if (endComp == 0 && startComp == seqValve)            return true;
      if (endComp == 0 && startComp == actuator)            return true;
      if (endComp == 0 && startComp == shuttleValve)       return true;
      
      if (startComp == seqValve && endComp == actuator)     return true;
      if (startComp == actuator && endComp == seqValve)     return true;
      if (startComp == priorityValve && endComp == actuator) return true;
      if (startComp == actuator && endComp == priorityValve) return true;
      if (startComp == actuator && endComp == shuttleValve) return true;
      if (startComp == shuttleValve && endComp == actuator) return true;
      
      if ((hose[i].getEndComp() == closedCenterValve) && 
         ((hose[i].getStartComp() == tee) || (hose[i].getStartComp() == shuttleValve) ||
         (hose[i].getStartComp() == priorityValve) || (hose[i].getStartComp() == actuator) ||
         (hose[i].getStartComp() == seqValve))){
         compTee = hose[i].getStartPortNum()/4;
         startComp = hose[i].getEndCompNum();
         if (checkTeeCompNum(compTee, startComp, startComp, true))
            result = true;
         }
          
         if (!result){
             if ((hose[i].getStartComp() == closedCenterValve) && 
             ((hose[i].getEndComp() == 0) || (hose[i].getEndComp() == shuttleValve) ||
             (hose[i].getEndComp() == priorityValve) || (hose[i].getEndComp() == actuator) ||
             (hose[i].getEndComp() == seqValve))){
             compTee = hose[i].getEndPortNum()/4;
             startComp = hose[i].getStartCompNum();
             if (checkTeeCompNum(compTee, startComp, startComp, true))
                  result = true;
           }
         
  }
  
        if (!result){
               if ((hose[i].getStartComp() == openCenterValve) && 
               ((hose[i].getEndComp() == tee) || (hose[i].getEndComp() == shuttleValve) ||
               (hose[i].getEndComp() == priorityValve) || (hose[i].getEndComp() == actuator) ||
               (hose[i].getEndComp() == seqValve))){
               compTee = hose[i].getEndPortNum()/4;
               startComp = hose[i].getStartCompNum();
               if (checkTeeCompNum(compTee, startComp, startComp, true))
                    result = true;
             }
                 
          if (!result){
             if ((hose[i].getEndComp() == openCenterValve) && 
             ((hose[i].getStartComp() == tee) || (hose[i].getStartComp() == shuttleValve) ||
             (hose[i].getStartComp() == priorityValve) || (hose[i].getStartComp() == actuator) ||
             (hose[i].getStartComp() == seqValve))){
             compTee = hose[i].getStartPortNum()/4;
             startComp = hose[i].getEndCompNum();
             if (checkTeeCompNum(compTee, startComp, startComp, true))
                  result = true;
           }
        }   
     }
  
  return result;
}


void resetTeeCount(){
  teeCount=0;
  for (int i=0;i<50;i++)
      teesChecked[i] = -1;
}

  // check connection of components through tees
boolean checkTeeHoses(int compTee, int endComp, int endPort, boolean sameType) // initCount = 1 for non-self to self check, 0 for self to self
{
   int maxCount;
   if (sameType) maxCount=2; else maxCount = 1;
   int currentCount = teeCount; 
   int hoses, tees;
   for(tees=0;tees<currentCount;tees++)                      // See if current tee already checked
   {
     if (teesChecked[tees]==compTee)
     {
       if (compTee >= 185){
         for (int tees2=tees;tees < currentCount;tees2++)
         {
          if (teesChecked[tees2]==compTee)
            return false;
         }
       if (compTee<185) 
          return false; // yes, exit                            // false if already checked
       } 
     }
   }
   
   teesChecked[teeCount++] = compTee;           // no, add to list
   
   for(hoses=0;hoses<numHoses;hoses++)
   {
     if(hose[hoses].checkCompNum(compTee))
     {
       if (hose[hoses].checkCompPort(endComp, endPort))  // is does this hose end where we are looking?
       {
         connCount++;
         if (connCount == maxCount) 
           return true;
       }
       // if another Tee, filter or standPipe, check again
       if (hose[hoses].getStartComp() == tee || hose[hoses].getStartComp() == filter || hose[hoses].getStartComp() == Standpipe) 
         if (checkTeeHoses(hose[hoses].getStartCompNum(),endComp, endPort,sameType)) return true;
       if (hose[hoses].getEndComp() == tee || hose[hoses].getEndComp() == filter || hose[hoses].getEndComp() == Standpipe) 
         if (checkTeeHoses(hose[hoses].getEndCompNum(),endComp, endPort, sameType)) return true;
     }
   }
  return false;                // if not found in path     
} // end checkTeeHoses

 // check if Tee connects to desired component on both ends
boolean checkTeeComp(int compTee, int startComp, int endComp)
{
    println(compTee); 
  int currentCount = teeCount; 
   int hoses, tees;
   for(tees=0;tees<currentCount;tees++)                      // See if current tee already checked
   {
     if (teesChecked[tees]==compTee)
     {
       if (compTee >= 185){
         for (int tees2=tees;tees < currentCount;tees2++)
         {
          if (teesChecked[tees2]==compTee)
            return false;
         }
       }
       if (compTee < 185) 
          return false; // yes, exit                            // false if already checked 
     }
   }

   teesChecked[teeCount++] = compTee;           // no, add to list
   
   for(hoses=0;hoses<numHoses;hoses++)
   {
     if(hose[hoses].checkCompNum(compTee))        // see if hose from number of tee connects to desired comp
     {
       if (hose[hoses].endComp == startComp || hose[hoses].startComp==startComp)
       {
         if (startComp != compLast){
           compCount++;
           compLast = startComp;
         }
       }
 
      else if (hose[hoses].endComp == endComp || hose[hoses].startComp==endComp)
       {
         if (endComp != compLast){
           compCount++;
           compLast = endComp;
         }
       }
       
       if (compCount==2) return true;       // if connects on both end, hose is good
       if (hose[hoses].getStartComp() == tee || hose[hoses].getStartComp() == filter || hose[hoses].getStartComp() == Standpipe){
             if (checkTeeComp(hose[hoses].getStartCompNum(),startComp, endComp))
               return true;
       }
               
       if (hose[hoses].getEndComp() == tee || hose[hoses].getEndComp() == filter || hose[hoses].getEndComp() == filter ) {
           if (checkTeeComp(hose[hoses].getEndCompNum(),startComp, endComp))
             return true;
       }
     }
   }
   return false;
} // end checkTeeHoses

        // check connect by component number instead of type
boolean checkTeeCompNum(int compTee, int startComp, int endComp, boolean checkActuator)
{
   int currentCount = teeCount; 
   int hoses, tees, comp1, comp2;
   for(tees=0;tees<currentCount;tees++)                      // See if current tee already checked
   {
     if (teesChecked[tees]==compTee)
     {
       return false; // yes, exit
     } 
   } 
   teesChecked[teeCount++] = compTee;           // no, add to list
   for(hoses=0;hoses<numHoses;hoses++)
   {
     if(hose[hoses].checkCompNum(compTee))      // check each hose, use port numbers/4 to get comp number
     {
       comp1 = hose[hoses].getStartPortNum()/4;
       comp2= hose[hoses].getEndPortNum()/4;
       if (comp1 == startComp || comp2==startComp)
       {
         compCount++;
       }
 
      else if (comp1 == endComp || comp2==endComp)
       {
         compCount++;
       }
       
       if (compCount==2) return true;
       
       if (hose[hoses].getStartComp() == tee || hose[hoses].getStartComp() == filter || hose[hoses].getStartComp() == Standpipe){
          if(checkActuator){
            if (checkTeeCompNum(hose[hoses].getStartPortNum()/4,startComp, endComp, true))
            return true;
          }
          if(!checkActuator){
            if (checkTeeCompNum(hose[hoses].getStartPortNum()/4,startComp, endComp, false))
               return true;
          }
       }
               
       if (hose[hoses].getEndComp() == tee || hose[hoses].getEndComp() == filter || hose[hoses].getEndComp() == Standpipe ) {
          if(checkActuator){
            if (checkTeeCompNum(hose[hoses].getEndPortNum()/4,startComp, endComp, true))
            return true;
          }
          if(!checkActuator){
            if (checkTeeCompNum(hose[hoses].getEndPortNum()/4,startComp, endComp, false))
               return true;
          }       
        }
       if (checkActuator){
             if (hose[hoses].getStartComp() == priorityValve || hose[hoses].getStartComp() == seqValve || hose[hoses].getStartComp() == shuttleValve || hose[hoses].getStartComp() == actuator){
                 if (checkTeeCompNum(hose[hoses].getStartPortNum()/4,startComp, endComp, true))
                  return true;
             }
               
             if (hose[hoses].getEndComp() == priorityValve || hose[hoses].getEndComp() == seqValve || hose[hoses].getEndComp() == shuttleValve || hose[hoses].getEndComp() == actuator ) {
                   if (checkTeeCompNum(hose[hoses].getEndPortNum()/4,startComp, endComp, true))
                   return true;
                   }
       }
     }
   }
   return false;
}  // end checkteecompnum

boolean testTee(int hoseNum){   // connection between Tee's
  boolean result = false;
       if (hose[hoseNum].getStartComp() == tee && hose[hoseNum].getEndComp()==0)
           result = true;
       if (hose[hoseNum].getStartComp() == filter && hose[hoseNum].getEndComp()==0)
           result = true;
       if (hose[hoseNum].getStartComp() == tee && hose[hoseNum].getEndComp()==filter)
           result = true;
         if (hose[hoseNum].getStartComp() == tee && hose[hoseNum].getEndComp()==Standpipe)
           result = true;
         if (hose[hoseNum].getStartComp() == Standpipe && hose[hoseNum].getEndComp()==0)
           result = true;
         if (hose[hoseNum].getStartComp() == Standpipe && hose[hoseNum].getEndComp()==filter)
           result = true;
         if (hose[hoseNum].getStartComp() == filter && hose[hoseNum].getEndComp()==Standpipe)
           result = true;
       return result;
     }
     
     // test if connection between given devices, ports, through ports.
boolean testIfConnection(int comp1,int port1, int comp2, int port2){
    boolean result = false;
    for(int i=0;i<numHoses;i++){
        if (testHose(i,comp1, port1, comp2, port2)) result = true;
    }
    return result;
}

void checkToggle(){            // see if switch was toggled
    if (mouseX>toggleX &&
      mouseY>toggleY &&
      mouseX<toggleX+toggleOn.width && 
      mouseY<toggleY+toggleOn.height){
       
        if (toggleFlag){
        toggleFlag= false;

        }
        else{
        toggleFlag= true;
        toggleCount++;
        }
      }
}
